import numpy as np, os, logging

logger = logging.getLogger(__name__)

from confapp import conf

try:
    
    import local_settings
    conf += local_settings
except:
    logger.info("Local settings file not available.")

from pyforms.basewidget import BaseWidget
from pyforms.controls import ControlText
from pyforms.controls import ControlCheckBox
from pyforms.controls import ControlFile
from pyforms.controls import ControlList
from pyforms.controls import ControlBoundingSlider

from pyforms.controls import ControlNumber
from pyforms.controls import ControlProgress

from idtrackerai.utils.segmentation_utils import cumpute_background
from idtrackerai.utils.py_utils import  getExistentFiles
from idtrackerai.constants import PROCESSES

from idtrackerai.list_of_fragments            import ListOfFragments
from idtrackerai.list_of_global_fragments     import ListOfGlobalFragments

from idtrackerai.video import Video

from idtrackerai.tracker_api import TrackerAPI
from idtrackerai.preprocessing_preview_api import PreprocessingPreviewAPI

from .gui.roi_selection import ROISelectionWin


from .helpers import Chosen_Video

import tensorflow as tf
with tf.Session() as sess:
    logger.info( "TENSORFLOW DEVICES: "+str(sess.list_devices()) )





class BaseIdTrackerAi(BaseWidget, ROISelectionWin):

    ERROR_MESSAGE_DEFAULT = "\n \nIf this error persists please open an issue at " \
                            "https://gitlab.com/polavieja_lab/idtrackerai or " \
                            "send an email to idtrackerai@gmail.com. " \
                            "Check the log file idtrackerai-app.log in your " \
                            "working directory and attach it to the issue."


    def __init__(self, *args, **kwargs):
        BaseWidget.__init__(self, title='idtracker.ai')
        ROISelectionWin.__init__(self)

        self._session    = ControlText('Session', default='test')
        self._video      = ControlFile('Video')
        self._video_path = ControlFile('Video file. Note: overwrite the _video parameter defined in the json. Nice to have to execute the application in a cluster environment')
        self._applyroi   = ControlCheckBox('Apply ROI', enabled=False)

        self._roi       = ControlList('ROI', enabled=False, readonly=True, select_entire_row=True,
                                      item_selection_changed_event=self.roi_selection_changed_evt,
                                      remove_function=self.remove_roi)

        self._bgsub     = ControlCheckBox('Subtract background', changed_event=self.bgsub_changed_evt, enabled=False)
        self._chcksegm  = ControlCheckBox('Check segmentation', enabled=False)
        self._resreduct = ControlNumber('Resolution reduction', default=conf.RES_REDUCTION_DEFAULT, minimum=0.01, maximum=1, decimals=2, step=0.1, enabled=False)

        self._intensity = ControlBoundingSlider('Intensity thresholds', default=[conf.MIN_THRESHOLD_DEFAULT, conf.MAX_THRESHOLD_DEFAULT], min=conf.MIN_THRESHOLD, max=conf.MAX_THRESHOLD, enabled=False)
        self._area      = ControlBoundingSlider('Area thresholds', default=[conf.MIN_AREA_DEFAULT, conf.MAX_AREA_DEFAULT], min=conf.AREA_LOWER, max=conf.AREA_UPPER, enabled=False)
        self._range     = ControlBoundingSlider('Tracking interval', default=[0,508], min=0, max=255, enabled=False) ###TODO: Change max of frames range to the number of frames of the video
        self._nblobs    = ControlNumber('Number of animals', default=conf.NUMBER_OF_ANIMALS_DEFAULT, enabled=False)
        self._progress  = ControlProgress('Progress', enabled=False)

        self._rangelst  = ControlText( 'Tracking intervals', visible=False )
        self._multiple_range = ControlCheckBox('Multiple', enabled=False)

        self._no_ids = ControlCheckBox('Track without identities')

        self.formset = [
            ('_video','_session'),
            ('_range','_rangelst','_multiple_range'),
            '_intensity',
            '_area',
            ('_nblobs', '_resreduct', ' ', '_applyroi', '_chcksegm', '_bgsub'),
            '_roi',
            ('_no_ids', '_progress')
        ]

        self.load_order = [
            '_session', '_video', '_range','_rangelst', '_multiple_range',
            '_intensity', '_area', '_nblobs',
            '_resreduct', '_chcksegm', '_roi', '_no_ids',
            '_bgsub'
        ]

        #self._video.value = '/home/ricardo/bitbucket/idtracker-project/idtrackerai_video_example.avi'

        self.bgsub_changed_evt()

        #self._resreduct.value = 0.3
        #self._area.value = [5, 50]

        # store the computed background with the correct resolution reduction
        self._background_img = None
        # store the computed background with the original size
        self._original_bkg = None
        # flag to open multiple video files with similar names
        self._multiple_files = False

    #########################################################
    ## GUI EVENTS ###########################################
    #########################################################

    def load_form(self, data, path=None):
        self._multiple_files = data.get('open-multiple-files', False)
        super().load_form(data, path)

    def bgsub_changed_evt(self):

        if self._bgsub.value:
            if self.video_path:
                video = Video( video_path=self.video_path, open_multiple_files=self.open_multiple_files )
                video.get_info()
                video._subtract_bkg = True
                video._original_bkg = cumpute_background(video)
                self._original_bkg = video.original_bkg
                video.resolution_reduction = self._resreduct.value
                self._background_img = video.bkg
            else:
                self._bgsub.value = False

    def set_controls_enabled(self, status):
        self._applyroi.enabled = status
        self._roi.enabled = status
        self._bgsub.enabled = status
        self._chcksegm.enabled = status
        self._resreduct.enabled = status
        self._intensity.enabled = status
        self._area.enabled = status
        self._range.enabled = status
        self._nblobs.enabled = status
        self._progress.enabled = status
        self._multiple_range.enabled = status
        self._rangelst.enabled = status
        self._no_ids.enabled = status



    def track_video(self):

        self._video.enabled=False
        self._session.enabled=False
        self.set_controls_enabled(False)

        try:
            if self.step1_pre_processing():
                if self._no_ids.value:
                    self.step2_wo_tracking()
                else:
                    self.step2_tracking()
        except Exception as e:
            logger.error(e, exc_info=True)
            self.critical( str(e), 'Error' )

        self._video.enabled=True
        self._session.enabled=True
        self.set_controls_enabled(True)



    def step1_pre_processing(self):

        if not os.path.exists(self.video_path):
            raise Exception("The video you are trying to track does not exist or the path to the video is wrong.")

        video_object = Video(
            video_path=self.video_path,
            open_multiple_files = self.open_multiple_files
        )
        video_object.get_info()

        # define the thresholds ranges
        video_object._min_threshold = self._intensity.value[0]
        video_object._max_threshold = self._intensity.value[1]

        # define the areas range
        video_object._min_area = self._area.value[0]
        video_object._max_area = self._area.value[1]

        # define the video range
        if self._multiple_range.value and self._rangelst.value:
            try:
                video_object._tracking_interval = eval(self._rangelst.value)
            except Exception as e:
                logger.fatal(e, exc_info=True)
                video_object._tracking_interval = [self._range.value]
        else:
            video_object._tracking_interval = [self._range.value]

        video_object._video_folder      = os.path.dirname(self.video_path)
        video_object._subtract_bkg      = self._bgsub.value
        video_object._original_bkg      = self._original_bkg
        video_object._number_of_animals = int(self._nblobs.value)
        video_object._apply_ROI         = self._applyroi.value
        video_object._original_ROI      = self.create_mask(video_object.original_height, video_object.original_width)

        video_object.resolution_reduction = self._resreduct.value

        #Check if it is identity transfer or knowledge_transfer
        if conf.IDENTITY_TRANSFER:
            video_object.check_and_set_identity_transfer_if_possible()

        video_object.create_session_folder(self._session.value)

        logger.debug("create Chosen_Video")
        chosen_video = Chosen_Video(video=video_object)
        logger.debug("before init PreprocessingPreviewAPI")
        pre = PreprocessingPreviewAPI( chosen_video )

        logger.debug("pre object: "+str(pre))

        self._progress.max = 9

        logger.debug('call: init_preview')
        pre.init_preview()
        self._progress.value = 1

        logger.debug('call: init_preproc_parameters')
        pre.init_preproc_parameters()
        self._progress.value = 2

        logger.debug('call: segment')
        pre.segment(
            self._intensity.value[0],
            self._intensity.value[1],
            self._area.value[0],
            self._area.value[1]
        )
        self._progress.value = 3

        video_object.create_images_folders() # for ram optimization
        logger.debug('call: compute_list_of_blobs')
        pre.compute_list_of_blobs()
        self._progress.value = 4

        logger.debug('call: check_segmentation_consistency')

        if self._chcksegm.value and \
           not pre.check_segmentation_consistency():
            outfile_path = os.path.join(video_object.session_folder, 'inconsistent_frames.csv')

            self.warning(
                "On some frames it was found more blobs than animals, "
                "you can find the index of these frames in the file:"
                "<p>{0}</p>"
                "<p>Please readjust the segmentation parameters and press 'Track video' again.</p>".format(outfile_path),
                'Found more blobs than animals'
            )
            with open(outfile_path, 'w') as outfile:
                outfile.write(
                    '\n'.join(map(str, pre.chosen_video.video.frames_with_more_blobs_than_animals))
                )

            self._progress.value = 0
            self._final_message = "Readjust the segmentation parameters and track the video again."
            return False

        self._progress.value = 5

        logger.debug('call: save_list_of_blobs_segmented')
        pre.save_list_of_blobs_segmented()
        self._progress.value = 6

        logger.debug('call: model_area_and_crossing_detector')
        pre.model_area_and_crossing_detector()
        self._progress.value = 7

        logger.debug('call: train_and_apply_crossing_detector')
        pre.train_and_apply_crossing_detector()
        self._progress.value = 8

        logger.debug('call: generate_list_of_fragments_and_global_fragments')
        pre.generate_list_of_fragments_and_global_fragments()
        self._progress.value = 9

        trainner = pre.crossing_detector_trainer
        list_of_fragments = pre.list_of_fragments
        list_of_global_fragments = pre.list_of_global_fragments

        return True

    def step2_tracking(self):

        video_folder   = os.path.dirname(self.video_path)
        session_folder = "session_{0}".format(self._session.value)

        videoobj_filepath   = os.path.join(video_folder, session_folder, 'video_object.npy')
        fragments_filepath  = os.path.join(video_folder, session_folder, 'preprocessing', 'fragments.npy')
        gfragments_filepath = os.path.join(video_folder, session_folder, 'preprocessing', 'global_fragments.npy')

        video_object = np.load(videoobj_filepath, allow_pickle=True).item()
        video_object.create_session_folder(self._session.value)

        if video_object.number_of_animals != 1:
            list_of_fragments        = ListOfFragments.load(fragments_filepath)
            list_of_global_fragments = ListOfGlobalFragments.load(gfragments_filepath, list_of_fragments.fragments)
        else:
            list_of_fragments        = None
            list_of_global_fragments = None

        chosen_video = Chosen_Video(
            video=video_object,
            list_of_fragments=list_of_fragments,
            list_of_global_fragments=list_of_global_fragments
        )

        chosen_video.existent_files, chosen_video.old_video = getExistentFiles(chosen_video.video, PROCESSES)

        tracker = TrackerAPI( chosen_video )

        tracker.start_tracking()

        if video_object.number_of_animals != 1:
            list_of_fragments.update_identification_images_dataset()

        logger.info("Estimated accuracy: {}".format(video_object.overall_P2))

        video_object.delete_data()

        self._final_message = "Tracking finished with {0:.2f} estimated accuracy.".format(video_object.overall_P2*100)


    def step2_wo_tracking(self):

        video_folder   = os.path.dirname(self.video_path)
        session_folder = "session_{0}".format(self._session.value)

        videoobj_filepath   = os.path.join(video_folder, session_folder, 'video_object.npy')
        fragments_filepath  = os.path.join(video_folder, session_folder, 'preprocessing', 'fragments.npy')
        gfragments_filepath = os.path.join(video_folder, session_folder, 'preprocessing', 'global_fragments.npy')

        video_object = np.load(videoobj_filepath, allow_pickle=True).item()
        video_object.create_session_folder(self._session.value)

        if video_object.number_of_animals != 1:
            list_of_fragments        = ListOfFragments.load(fragments_filepath)
            list_of_global_fragments = ListOfGlobalFragments.load(gfragments_filepath, list_of_fragments.fragments)
        else:
            list_of_fragments        = None
            list_of_global_fragments = None

        chosen_video = Chosen_Video(
            video=video_object,
            list_of_fragments=list_of_fragments,
            list_of_global_fragments=list_of_global_fragments
        )

        chosen_video.existent_files, chosen_video.old_video = getExistentFiles(chosen_video.video, PROCESSES)

        tracker = TrackerAPI( chosen_video )

        tracker.track_wo_identities()

        video_object.delete_data()

        self._final_message = "Tracking without identities finished. No estimated accuracy computed."



    def __update_progress(self, value, label=None, total=None):

        if total is not None: self._progress.max   = total
        if label is not None: self._progress.label = label

        self._progress.value = value

    def alert(self,     msg, title=None):
        msg = msg + self.ERROR_MESSAGE_DEFAULT
        self.message(msg, title, msg_type='error')

    def critical(self,  msg, title=None):
        msg = msg + self.ERROR_MESSAGE_DEFAULT
        self.message(msg, title, msg_type='error')

    @property
    def open_multiple_files(self):
        return self._multiple_files

    @property
    def video_path(self):
        return self._video_path.value if self._video_path.value else self._video.value



#!/usr/bin/python
# -*- coding: utf-8 -*-

import logging

from confapp import conf

from AnyQt.QtWidgets import QMdiArea, QMdiSubWindow

from pyforms_gui.controls.control_base import ControlBase

logger = logging.getLogger(__name__)


class ControlMdiArea(ControlBase, QMdiArea):
	"""
	The ControlMdiArea wraps a QMdiArea widget which provides
	 an area in which MDI windows are displayed.
	"""

	def __init__(self, *args, **kwargs):
		QMdiArea.__init__(self)
		ControlBase.__init__(self, *args, **kwargs)
		self._showCloseButton = True
		self._openWindows = []

		self.logger = logging.getLogger(__name__)

	def __sub__(self, widget):
		"""
		Remove subwindow and unassigned it from widget
		:param widget:
		:return:
		"""
		widget.close()
		self += widget  # little tweak to temporarily make this widget as the active subwindow
		self.removeSubWindow(widget.subwindow)
		del widget.subwindow

		logger.debug("Widget sub window removed. MDI area sub windows: %s", self.subWindowList())

		return self

	def __add__(self, widget):
		"""
		Show widget on mdi area.

		If widget does not have a subwindow assigned, create a new subwindow without enabling the WA_DeleteOnClose event.
		This will allow subwindow to be hidden instead of destroyed. Otherwise, the closeEvent.accept() will cause
		the "Internal c++ Object Already Deleted" problem.

		If widget already has a subwindow, just show them (both the subwindow and the widget inside)!
		:param widget:
		:return:
		"""

		if not hasattr(widget, 'subwindow'):
			subwindow = QMdiSubWindow()
			subwindow.setWidget(widget)
			rect = widget.geometry()
			# DO NOT SET ATTRIBUTE WA_DeleteOnClose because we want window not to be destroyed
			widget.subwindow = self.addSubWindow(subwindow)
			subwindow.setGeometry(rect)

		widget.subwindow.show()
		widget.show()
		widget.closeEvent = lambda x: self._subWindowClosed(x)
		widget.setFocus()

		logger.debug("Sub window opened. MDI area sub windows: %s", self.subWindowList())

		return self

	def _subWindowClosed(self, closeEvent):
		"""
		Perform actions when subwindow is closed.
		In this case, we don't want subwindow to be removed nor destroyed in order to reutilize later.
		The closeEvent.accept() will just hide the subwindow.
		:param closeEvent:
		:return:
		"""

		try:
			window = self.activeSubWindow()

			if window:
				widget = window.widget()
				# self.removeSubWindow(window) DO NOT REMOVE TO KEEP WINDOW POSITION
				widget.before_close_event()
			else:
				if hasattr(window, 'before_close_event'):
					widget.before_close_event()

			closeEvent.accept()

			logger.debug("Sub window closed. MDI area sub windows: %s", self.subWindowList())
		except Exception as err:
			logger.warning(str(err))


	##########################################################################
	############ Properties ##################################################
	##########################################################################

	@property
	def show_subwin_close_button(self): return self._showCloseButton

	@show_subwin_close_button.setter
	def show_subwin_close_button(self, value): self._showCloseButton = value

	@property
	def label(self): return self._label

	@label.setter
	def label(self, value): self._label = value

	@property
	def form(self): return self



# !/usr/bin/python
# -*- coding: utf-8 -*-
from AnyQt.QtGui import QColor
from PyQt5.QtCore import QPoint


class Graph(object):
	"""
	"""

	def __init__(self, timeLineWidget, color=QColor(255, 0, 0), name='undefined'):
		"""		
		:param timeLineWidget: 
		:param color: 
		:param name: 
		"""
		self._data 		= []
		self._graph_max = None
		self._graph_min = None
		self._widget 	= timeLineWidget
		self._color 	= color
		self._zoom 		= 1.0
		self._top 		= 0
		self._name 		= name

		timeLineWidget += self

	def __unicode__(self): return self._name
	def __str__(self):     return self._name

	def __len__(self): 					 return len(self._data)
	def __getitem__(self, index):
		return self._data[index] if index<len(self) else None
	def __setitem__(self, index, value): 
		if index >= len(self):
			for i in range(len(self), index + 1): self._data.append(None)

		if value is not None:
			if value > self._graph_max: self._graph_max = value
			if value < self._graph_min: self._graph_min = value

		if index is None or value is None: 
			self._data[index] = None
		else:
			self._data[index] = value
	
	def import_data(self, data):
		"""
		Import the data from an array
		:param list(tuple(int,float)) data: List of coordinates (frame, value).
		"""
		self._graph_max = 0
		self._graph_min = 100000000000
		self._data 		= []
		for x, y in data:
			self[int(x)] = float(y)


	def remove(self):
		"""
		Remove the graph from the timeline
		"""
		self._widget -= self


	#####################################################################################
	###### PROPERTIES ###################################################################
	#####################################################################################

	@property
	def graph_min(self):
		return self._graph_min

	@graph_min.setter
	def graph_min(self, value):
		self._graph_min = value

	@property
	def graph_max(self):
		return self._graph_max

	@graph_max.setter
	def graph_max(self, value):
		self._graph_max = value

	@property
	def zoom(self):
		return self._zoom

	@zoom.setter
	def zoom(self, value):
		self._zoom = value

	@property
	def top(self):
		return self._top

	@top.setter
	def top(self, value):
		self._top = value

	#####################################################################################
	#####################################################################################
	#####################################################################################

	def draw(self, painter, left, right, top, bottom):
		"""
		
		:param painter: 
		:param left: 
		:param right: 
		:param top: 
		:param bottom: 
		:return: 
		"""
		painter.setPen(self._color)
		painter.setOpacity(0.7)

		fov_height 	 = (bottom - top) * self.zoom  #calculate the height visible
		start 		 = self._widget.x2frame(left)  #calculate the start frame to draw
		end 		 = self._widget.x2frame(right) #calculate the end frame to draw
		end 		 = len(self) if end > len(self) else end #check if the end frame his higher than the available data
		diff_max_min = (self._graph_max - self._graph_min) #calculate the difference bettween the lower and higher value

		# in case the frames have always de same value
		# set artificially a min and max to plot values at the middle
		if diff_max_min == 0:
			self._graph_min = self._graph_max - 1
			self._graph_max = self._graph_max + 1
			diff_max_min = 2
		elif diff_max_min < 0:
			diff_max_min = 1

		top = (-self._graph_min if self._graph_min > 0 else abs(self._graph_min)) * self._zoom

		last_coordinate   = None
		last_real_x_coord = None

		data_len_minus1 = len(self)-1

		for x in range(start, end+1):
			y = self[x]
			if y is None:
				last_coordinate = None
				last_real_x_coord = None
			else:
				y_pixel = self._top + ((top + y) * fov_height) // diff_max_min

				if 0<x<data_len_minus1 and self[x-1] is None and self[x+1] is None:
					painter.drawEllipse(
						QPoint(self._widget.frame2x(x), fov_height - y_pixel),
						2, 2
					)
				else:
					if last_coordinate:
						diff_frames = abs(x - last_real_x_coord)
						draw_from_coord = last_coordinate if diff_frames == 1 else (self._widget.frame2x(x), fov_height - y_pixel)
						painter.drawLine(draw_from_coord[0], draw_from_coord[1], self._widget.frame2x(x), fov_height - y_pixel)

					last_coordinate = self._widget.frame2x(x), fov_height - y_pixel
					last_real_x_coord = x


		painter.setOpacity(1.0)

	@property
	def name(self): return self._name

	@name.setter
	def name(self, value):
		if value != self._name:
			self._name = value

			if not hasattr(self, 'renaming_graph_flag'):
				self.renaming_graph_flag = True
				i = self._widget.graphs.index(self)
				self._widget.control.rename_graph(i, value)
				del self.renaming_graph_flag
		

	def mouse_move_evt(self, event, top, bottom):
		"""
		:param event: 
		:param top: 
		:param bottom: 
		:return: 
		"""

		frame 			= self._widget.x2frame(event.x())
		fov_height 		= (bottom - top) * self._zoom
		top 			= (-self._graph_min if self._graph_min > 0 else abs(self._graph_min)) * self._zoom
		diff_max_min 	= (self._graph_max - self._graph_min)
		if diff_max_min <= 0: diff_max_min = 1

		# no value
		if self[frame] is None: 
			self._widget.graphs_properties.coordinate_text = None
		else:
			self._widget.graphs_properties.coordinate_text = "Frame: {0} Value: {1}".format(frame, self[frame])


	def export_2_file(self, filename):
		"""
		
		:param filename: 
		:return: 
		"""
		with open(filename, 'w') as outfile:
			outfile.write(';'.join(['frame', 'value']) + '\n')
			for x, y in enumerate(self._data):
				if y is not None: outfile.write(';'.join([str(x), str(y)]) + '\n')



# !/usr/bin/python
# -*- coding: utf-8 -*-

from confapp import conf

from AnyQt.QtWidgets import QDialog, QInputDialog, QColorDialog
from AnyQt.QtGui import QColor, QPixmap, QFont, QPainter
from AnyQt import uic
from AnyQt import QtCore


import pyforms_gui.utils.tools as tools


class TimelinePopupWindow(QDialog):
	"""
	Opens a dialog where the user can specify the behavior annotated
	in the selected line, as well as some related options.

	The parent timeline widget must be given, as well as the track
	identifier to edit.
	"""

	def __init__(self, parent, track_id):
		"""
		
		:param parent: 
		:param track_id: 
		"""
		super(TimelinePopupWindow, self).__init__(parent=parent)
		self._parent = parent
		control_path = tools.getFileInSameDirectory(
			__file__, "win_track.ui")
		self._ui = uic.loadUi(control_path)
		self._ui.setWindowTitle("Row {:d} properties".format(track_id + 1))

		# Dialog variables
		self.behaviors = []
		self.behavior = None
		self.color = self._parent.color
		self.current_track = track_id

		self._default_comboBox_text = "Add a new label"
		self.__get_existing_tracklabels()

		# Set default color display
		self.__preview_color()

		# SIGNALS
		self._ui.comboBox.currentIndexChanged.connect(
			self.__on_comboBox_change)
		self._ui.pushButton_add.clicked.connect(self.__add_behavior)
		self._ui.pushButton_remove.clicked.connect(self.__remove_behavior)
		self._ui.pushButton_color.clicked.connect(self.__pick_color)

	def __on_comboBox_change(self):
		"""
		Handles comboBox index change.
		"""
		cb = self._ui.comboBox
		self.behavior = cb.itemText(cb.currentIndex())

	def __add_behavior(self):
		"""
		Add a behavior to the already existing ones.
		"""
		cb = self._ui.comboBox
		text, ok = QInputDialog.getText(
			self, 'Add behavior', 'Description:', text='')
		if ok:
			self.behavior = str(text)
			self.behaviors.append(self.behavior)
			self._ui.comboBox.addItem(self.behavior)
			cb.setCurrentIndex(cb.findText(self.behavior))

		# If adding the first item, we need to enable the comboBox and
		# remove the placeholder text
		if not cb.isEnabled():
			cb.removeItem(cb.findText(self._default_comboBox_text))
			cb.setEnabled(True)

	def __remove_behavior(self):
		"""Remove a behavior from the already existing ones."""
		cb = self._ui.comboBox
		i = cb.currentIndex()
		self.behaviors.remove(str(cb.itemText(i)))
		cb.removeItem(i)

		# If there are no behaviors assigned, just fill the comboBox
		# with a placeholder
		if cb.count() < 1:
			cb.addItem(self._default_comboBox_text)
			cb.setEnabled(False)

	def __pick_color(self):
		"""Dialog to choose a color."""
		self.color = QColorDialog.getColor(self.color)
		self.__preview_color()

	def __preview_color(self):
		"""
		Shows selected colors in two QLabel widgets.

		The first shows true color, while the second presents the color
		with an opacity as seen in the timeline and with some dummy text
		to preview readability.
		"""
		pixmap = QPixmap(50, 25)
		color = QColor(*self.color.getRgb())

		# Preview color
		color.setAlpha(int(255 * 1.0))
		pixmap.fill(color)
		self._ui.label_color.setPixmap(pixmap)

		# Preview color with transparency and some text
		color.setAlpha(int(255 * 0.5))
		pixmap.fill(color)
		painter = QPainter(pixmap)
		painter.setFont(QFont('Decorative', 8))
		painter.drawText(pixmap.rect(), QtCore.Qt.AlignCenter, "Text")
		painter.end()
		self._ui.label_color_alpha.setPixmap(pixmap)

	def __get_existing_tracklabels(self):
		"""
		Gets existing track labels already assigned.

		Scans the timeline track labels and sets the comboBox value
		accordingly.
		"""
		cb = self._ui.comboBox

		# Loop across timeline labels
		for index, track in enumerate(self._parent._tracks):
			# If there is already an assigned label, append it to the
			# behaviors list and to the comboBox (if not a duplicate)
			if track.title != '':
				self.behavior = track.title
				if self.behavior not in self.behaviors:
					cb.addItem(self.behavior)
					self.behaviors.append(self.behavior)

				# Set comboBox value to the one of the selected track
				if index == self.current_track:
					cb.setCurrentIndex(cb.findText(self.behavior))

		# If there are no behaviors assigned yet, just fill the comboBox
		# with a placeholder
		if cb.count() < 1:
			cb.addItem(self._default_comboBox_text)
			cb.setEnabled(False)

########################################################################
###
# MAYBE IN THE FUTURE IMPLEMENT THIS USING THE BaseWidget

# class TimelinePopupWindow(BaseWidget):
#     """
#     Opens a dialog where the user can specify the behavior annotaded
#     in the selected line, as well as some related options.
#     """

#     def __init__(self, parent=None):
#         super(TimelinePopupWindow, self).__init__('Adjust line options')

#         if parent is not None:
#             self.parent = parent

#         self._behaviors = None
#         self._behaviorname  = None
#         self._linecolor = None

#         # Combobox
#         self._combobox = ControlCombo()

#         # Buttons
#         self._btn_add = ControlButton()
#         self._btn_remove = ControlButton()
#         self._btn_import = ControlButton()
#         self._btn_export = ControlButton()
#         self._btn_color = ControlButton()
#         self._btn_ok = ControlButton()
#         self._btn_cancel = ControlButton()

#         self._formset = ['_combobox',
#                          ('_btn_add', '_btn_remove', '_btn_import', '_btn_export'),
#                          ('_btn_ok', '_btn_cancel')]

#         def __choose_color(self):
#             pass
#             QColorDialog.getColor()



import csv
from pyforms_gui.basewidget import BaseWidget
from pyforms_gui.controls.control_progress import ControlProgress
from pyforms_gui.controls.control_button import ControlButton
from pyforms_gui.controls.control_list import ControlList
from pyforms_gui.controls.control_number import ControlNumber
from pyforms_gui.controls.control_slider import ControlSlider
from pyforms_gui.controls.control_textarea import ControlTextArea
from pyforms_gui.controls.control_text import ControlText
from pyforms_gui.controls.control_number import ControlNumber
from pyforms_gui.controls.control_label import ControlLabel
from pyforms_gui.controls.control_combo import ControlCombo
from pyforms_gui.controls.control_emptywidget import ControlEmptyWidget
from pyforms_gui.controls.control_file import ControlFile

        
import time
import datetime
import dateutil.parser

import logging

logger=logging.getLogger(__file__)

class BonsaiImportFileDlg(BaseWidget):

    def __init__(self, timeline=None):
        super(BonsaiImportFileDlg, self).__init__('Import file')

        self._file          = ControlFile('File to import')
        self._fps           = ControlNumber('Video FPS', default=30)
        self._startframe    = ControlNumber('Initial frame', visible=False, maximum=100000)


        self._formset = [('_startframe','_fps', '_file')]

        #self._file.value = '/home/ricardo/Desktop/rat1e2.csv'


class ImportWindow(BaseWidget):

    def __init__(self, control):
        super(ImportWindow, self).__init__('Import file', parent_win=control.parent)
        self.setContentsMargins(10, 10, 10, 10)
        self._timeline = control

        # Definition of the forms fields
        self._filetype      = ControlCombo('Please select the type of file you would like to import:')
        self._importButton  = ControlButton('Import')
        self._panel         = ControlEmptyWidget('Panel')
        self._file          = ControlFile('File to import')
        
        self._panel.value = self._file
        self._filetype.add_item('Events file', 0)
        self._filetype.add_item('Graph file', 1)
        self._filetype.add_item('Bonsai events file', 2)
        self._filetype.add_item('Bonsai events file (old format)', 3)

        self._formset = [
            ('_filetype', ' '),
            '_panel',
            (' ', '_importButton'),
            ' '
        ]

        self._filetype.changed_event = self.__fileTypeChanged
        self._importButton.value = self.__importData

        from pyforms_gui.dialogs.csv_parser import CsvParserDialog
        self._graphCsvParserDlg = CsvParserDialog(control)
        self._graphCsvParserDlg.xField.label = "Value column"
        self._graphCsvParserDlg.yField.hide()
        self._graphCsvParserDlg.zField.hide()
        self._graphCsvParserDlg.loadButton.hide()

        self._bonsai_import_dlg = BonsaiImportFileDlg()

    def __fileTypeChanged(self):
        if self._filetype.value == 0:
            self._panel.value = self._file

        elif self._filetype.value == 1:
            self._panel.value = self._graphCsvParserDlg

        elif self._filetype.value == 2:
            self._panel.value = self._bonsai_import_dlg
            self._bonsai_import_dlg._startframe.hide()

        elif self._filetype.value == 3:
            self._panel.value = self._bonsai_import_dlg
            self._bonsai_import_dlg._startframe.show()













    def __importData(self):
        if not self._file.value: return

        if self._filetype.value == 0:
            separator = ','
            with open(self._file.value, 'rU') as csvfile:
                line = csvfile.readline()
                if ";" in line:
                    separator = ';'
            with open(self._file.value, 'rU') as csvfile:
                csvfile = csv.reader(csvfile, delimiter=separator)
                self._timeline._time.import_events_from_csvreader(csvfile)

        elif self._filetype.value == 1:
            self._timeline._time.add_graph('Untitled graph', self._graphCsvParserDlg)
            self._timeline.open_graphs_properties()

        elif self._filetype.value == 2:
            self.__import_bonsai_events()
        elif self._filetype.value == 3:
            self.__import_bonsai_events_oldformat()

        self._timeline.repaint()
        self._timeline._time.repaint()
        self.close()  # pylint: disable=no-member


    def import_chart(self, filename, frame_col=0, val_col=1):
        self._filetype.value = 1            
        self._graphCsvParserDlg.filename    = filename
        self._graphCsvParserDlg.frameColumn = frame_col
        self._graphCsvParserDlg.xColumn     = val_col
        
            

    def __import_bonsai_events(self):
        
        with open(self._bonsai_import_dlg._file.value, 'rU') as csvfile:
            values = []
            pointEventValues = []
            csvfile = csv.reader(csvfile, delimiter=' ')
            for row in csvfile:  # strip Start/End word from all events names which are not PointEven
                
                try:
                    timestr = row[1].rstrip('0')
                    cvttime = datetime.datetime.strptime(timestr, "%H:%M:%S.%f")
                except:
                    timestr = row[1]
                    try:
                        cvttime = datetime.datetime.strptime(timestr, "%H:%M:%S")
                    except:
                        timestr = timestr.replace('T', ' ')
                        cvttime = dateutil.parser.parse(timestr)

                seconds = (cvttime - datetime.datetime(1900, 1, 1)).total_seconds()
                frame = int(round(self._bonsai_import_dlg._fps.value * seconds))

                if row[2] == "PointEvent":
                    eventtype = row[0]
                    pointEventValues.append([eventtype, frame, row[2]])
                else:
                    if row[0].startswith('Start'):
                        eventtype = row[0][len('Start'):]  # strip Start word from the beginning
                    else:
                        eventtype = row[0][len('End'):]  # strip End word from the beginning
                    values.append([eventtype, frame, row[2]])

            values = sorted(values, key=lambda x: (x[0].capitalize(), x[1]))
            pointEventValues = sorted(pointEventValues, key=lambda x: (x[0].capitalize(), x[1]))
            ntracks = len(set([x[0] for x in values])) + 1

            # collapse
            events = []
            eventsTypes = {}  # Events names
            currentTrack = 0
            for index in range(0, len(pointEventValues)):
                pointEventValue = pointEventValues[index]
                eventsTypes[pointEventValue[0]] = currentTrack
                self._timeline.add_event(pointEventValue[1], pointEventValue[1] + 50, title=pointEventValue[0], row=currentTrack)

            currentTrack = 1

            for index in range(0, len(values), 2):
                row0 = values[index]
                row1 = values[index + 1]

                if row0[0] not in eventsTypes:
                    eventsTypes[row0[0]] = currentTrack
                    track = currentTrack
                    currentTrack += 1
                else:
                    track = eventsTypes[row0[0]]

                self._timeline.add_event(row0[1], row1[1], title=row0[0], row=track)




    def __import_bonsai_events_oldformat(self):
        
        try:
            with open(self._bonsai_import_dlg._file.value, 'rU') as csvfile:
                windows_events = []
                points_events  = []
                first_date     = None

                initial_frame = int(self._bonsai_import_dlg._startframe.value)

                open_windows = {}
                for row in csvfile:
                    row = row[:-1] #remove the newline character

                    if row.endswith('WindowClosing'):
                        split     = row.rfind(' ')
                        #eventtype = row[-split:]
                        eventtype = 'end'
                        timestr   = row[split-33:split]
                        eventname = row[3:split-33].strip()
                    elif row.endswith('PointEvent'):
                        split     = row.rfind(' ')
                        eventtype = row[split+1:]
                        timestr   = row[split-33:split]
                        eventname = row[:split-33]
                    else:
                        eventtype = 'start'
                        timestr   = row[-33:]
                        eventname = row[5:-33].strip()
                    
                    cvttime = dateutil.parser.parse(timestr.replace('T', ' '))
                    cvttime = cvttime.replace(tzinfo=None)
                    if first_date is None: first_date = cvttime
                    
                    seconds = (cvttime - first_date).total_seconds()
                    frame   = int(round(self._bonsai_import_dlg._fps.value * seconds)) + initial_frame

                    if eventtype=='PointEvent':
                        points_events.append([eventtype, frame, eventname])
                    else:
                        if eventtype=='start':
                            open_windows[eventname] = frame
                        elif eventtype=='end':
                            begin_frame = open_windows[eventname]
                            del open_windows[eventname]
                            windows_events.append( [eventtype, (begin_frame, frame) , eventname] )
                    

                windows_events = sorted(windows_events, key=lambda x: (x[0].capitalize(), x[2], x[1]))
                points_events  = sorted(points_events,  key=lambda x: (x[0].capitalize(), x[1]))
                ntracks        = len(set([x[0][1] for x in windows_events])) + 1

                # collapse
                events = []
                events_types = {}  # Events names
                current_track = 0
                for index in range(0, len(points_events)):
                    eventtype, frame, eventname = points_events[index]
                    events_types[eventname] = current_track
                    self._timeline.add_event(frame, frame+2, title=eventname, row=current_track)

                current_track = 1

                for event in windows_events:
                    eventtype, (frame_begin, frame_end), eventname = event
                    
                    if eventname not in events_types:
                        events_types[eventname] = current_track
                        track         = current_track
                        current_track += 1
                    else:
                        track = events_types[eventname]

                    self._timeline.add_event(frame_begin, frame_end, title=eventname, row=track)


        except Exception as e:
            self.warning("An error occurred when trying to import the file. Please check the logs.")
            logger.error(e, exc_info=True)


#!/usr/bin/python
# -*- coding: utf-8 -*-

from confapp import conf

from AnyQt           import uic
from AnyQt.QtWidgets import QPushButton, QSizePolicy
from AnyQt.QtGui     import QIcon

from pyforms_gui.controls.control_base import ControlBase


class ControlButton(ControlBase):
    def __init__(self, *args, **kwargs):
        """
        ...

        :param str icon: Button icon
        :param bool checkable: Flag to set the button checkable.
        
        """
        self._checkable = kwargs.get('checkable', False)
        super(ControlButton, self).__init__(*args, **kwargs)

        default = kwargs.get('default', None)
        if default: self.value = default

        icon = kwargs.get('icon', None)
        if icon: self.icon = icon

    def init_form(self):
        self._form = QPushButton()
        self._form.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        self._form.setCheckable(self._checkable)
        self.label = self._label
        self._form.setToolTip(self.help)
        super(ControlButton, self).init_form()

    def click(self):
        """
        Trigger a click event
        """
        self._form.click()

    def load_form(self, data, path=None):
        pass

    def save_form(self, data, path=None):
        pass

    ##########################################################################

    @property
    def label(self):
        return ControlBase.label.fget(self)

    @label.setter
    def label(self, value):
        ControlBase.label.fset(self, value)
        self._form.setText(self._label)

    @property
    def icon(self):
        """
        Sets and gets the icon of the button.
        """
        return self._form.icon()

    @icon.setter
    def icon(self, value):
        if isinstance(value, (str, bytes)):
            self._form.setIcon(QIcon(value))
        else:
            self._form.setIcon(value)

    ##########################################################################

    @property
    def value(self):
        """
        Sets and gets the value of the Button. The value should be a function
        """
        return None

    @value.setter
    def value(self, value):
        try:
            self._form.clicked.disconnect()  # ignore previous signals if any
        except TypeError as err:
            # http://stackoverflow.com/questions/21586643/pyqt-widget-connect-and-disconnect
            pass
        self._form.clicked[bool].connect(value)

    @property
    def checked(self):
        """
        Sets and gets the button checked state
        """
        return self._form.isChecked()

    @checked.setter
    def checked(self, value):
        self._form.setChecked(value)



from pyforms.basewidget import BaseWidget
from pyforms.controls import ControlNumber
from pyforms.controls import ControlButton

class RangeWin(BaseWidget):

    def __init__(self, *args, **kwargs):
        super().__init__(title='Frames range')

        self._list = kwargs.get('control_list', None)

        self._begin = ControlNumber('Begin',
                        default=kwargs.get('begin', 0),
                        maximum=kwargs.get('maximum', 100)
                    )
        self._end   = ControlNumber('End',
                        default=kwargs.get('begin', 0),
                        maximum=kwargs.get('maximum', 100)
                    )
        self._okbtn = ControlButton('Ok', default=self.__okbtn_evt)

        self.formset = [
            ('_begin', '_end'),
            '_okbtn'
        ]

        self.set_margin(10)

    def __okbtn_evt(self):
        if self._begin.value>self._end.value:
            self.alert('The <b>begin</b> frame has to be smaller than <b>end</b> frame.')
        else:
            value = str( [int(self._begin.value), int(self._end.value)] )
            if self._list.value:
                self._list.value += ', '+value
            else:
                self._list.value = value
            self.close()



# !/usr/bin/python
# -*- coding: utf-8 -*-

import os

PYFORMS_MAIN_WINDOW_ICON_PATH = None

PYFORMS_MAINWINDOW_MARGIN = 7

PYFORMS_CONTROL_CODE_EDITOR_DEFAULT_FONT_SIZE = '12'
PYFORMS_CONTROL_EVENTS_GRAPH_DEFAULT_SCALE = 1

PYFORMS_QUALITY_TESTS_PATH = None

PYFORMS_STYLESHEET = None
PYFORMS_STYLESHEET_DARWIN = None
PYFORMS_STYLESHEET_LINUX = None
PYFORMS_STYLESHEET_WINDOWS = None

PYFORMS_CONTROLPLAYER_FONT = 9

# In a normal loading, there may be errors that show up which are not important.
# This happens because plugins_finder will search for classes on plugins which are not present because they are not needed.
# However, if plugin is not loaded at all, this will show all related errors.
# See pyforms_gui.utils.plugins_finder.find_class()
PYFORMS_SILENT_PLUGINS_FINDER = True

PYFORMS_QSCINTILLA_ENABLED 	= True
PYFORMS_MATPLOTLIB_ENABLED 	= True
PYFORMS_WEB_ENABLED 		= True
PYFORMS_GL_ENABLED 			= True
PYFORMS_VISVIS_ENABLED 		= True


PYFORMS_DIALOGS_OPTIONS = None
PYFORMS_COLORDIALOGS_OPTIONS = None

#from AnyQt.QtWidgets import QColorDialog, QFileDialog
#PYFORMS_DIALOGS_OPTIONS = QFileDialog.DontUseNativeDialog
#PYFORMS_COLORDIALOGS_OPTIONS = QColorDialog.DontUseNativeDialog

import logging
PYFORMS_LOG_HANDLER_LEVEL = logging.INFO
PYFORMS_LOG_FORMAT = '[%(levelname)-7s] %(name)-60s %(message)s'


#!/usr/bin/python
# -*- coding: utf-8 -*-

from pyforms_gui.controls.control_text import ControlText
from AnyQt.QtWidgets import QLineEdit

class ControlPassword(ControlText):

    def init_form(self):
        super(ControlPassword, self).init_form()
        
        self.form.label.setAccessibleName('ControlPassword-label')
        self.form.lineEdit.setEchoMode(QLineEdit.Password)


# !/usr/bin/python
# -*- coding: utf-8 -*-

from AnyQt.QtGui import QColor
from pyforms_gui.controls.control_events_graph.Track import Track


class Event(object):
	"""
	
	"""

	def __init__(self, begin, end=30, title=None, parentWidget=None, color='#FFFF00'):
		"""
		
		:param begin: 
		:param end: 
		:param title: 
		:param parentWidget: 
		:param color: 
		"""
		self._parentWidget = parentWidget
		self._title = title
		self._begin = begin
		self._end = end
		self.color = color

	def __repr__(self):
		return self._end

	def __gt__(self, other):
		return self._end > other

	def __lt__(self, other):
		return self._end < other

	def __ge__(self, other):
		return self._end >= other

	def __le__(self, other):
		return self._end <= other

	def __ne__(self, other):
		return self._end != other

	def __eq__(self, other):
		return self._end == other

	##########################################################################
	#### HELPERS/FUNCTIONS ###################################################
	##########################################################################

	def draw(self, painter, top=20, showvalues=True, left_shift=0, scale=1.0):
		"""
		
		:param painter: 
		:param top: 
		:param showvalues: 
		:param left_shift: 
		:param scale: 
		:return: 
		"""
		start, width = left_shift + self.begin / scale, self.end / scale - self.begin / scale

		painter.setPen(QColor(0, 0, 0))
		painter.setOpacity(0.5)
		painter.drawRect(start, top, width, self._parentWidget.tracks_height - 4)
		painter.setOpacity(1.0)

	# if showvalues:
	#	painter.drawText( start, top , "[{0};{1}] delta:{2}".format(self._begin, self._end, self._end - self._begin) )

	def remove(self):
		"""
		
		:return: 
		"""
		try:
			self._parentWidget._tracks[self.track].events.remove(self)
		except:
			pass

	##########################################################################
	#### PROPERTIES ##########################################################
	##########################################################################

	@property
	def title(self):
		return self._title

	@property
	def begin(self):
		return self._begin

	@begin.setter
	def begin(self, value):
		self._begin = value
		if self._begin < 0: self._begin = 0

	@property
	def end(self):
		return self._end

	@end.setter
	def end(self, value):
		self._end = value

	# if self._end > (self._parentWidget.width()): self._end = (self._parentWidget.width())

	@property
	def color(self):
		return self._defautcolor

	@color.setter
	def color(self, value):
		self._defautcolor = QColor(value) if (type(value) == str) else value



# !/usr/bin/python
# -*- coding: utf-8 -*-

from AnyQt.QtGui import QColor
from AnyQt import QtCore
from .event import Event


class Track(object):
	"""
	Track
	"""

	DEFAULT_COLOR = QColor(100, 100, 255)

	def __init__(self, widget, title='', color=None):
		self._title   = title
		self._color   = self.DEFAULT_COLOR if color is None else color
		self._widget  = widget
		self._index   = len(widget.tracks)
		self._events = []

	def __len__(self):
		return len(self._events)

	def __add__(self, other):
		if isinstance(other, Event):
			self._events.append(other)
			self._events = sorted(self._events, key=lambda e: e.begin)
		return self

	def __sub__(self, other):
		if isinstance(other, Event):
			self._events.remove(other)
		return self

	@staticmethod
	def whichTrack(y):
		return (y - 20) // 34

	@staticmethod
	def which_top(track):
		return track * 34 + 20

	@property
	def events(self):
		return self._events

	@property
	def color(self):
		return self._color

	@color.setter
	def color(self, value):
		self._color = value

	@property
	def title(self):
		return self._title

	@title.setter
	def title(self, value):
		self._title = value

	@property
	def index(self):
		"""
		:return: Index of the track in the timeline.
		"""
		return self._index

	@index.setter
	def index(self, value):
		self._index = value

	@property
	def events(self):
		return self._events

	def remove(self):
		"""
		Remove the track
		"""
		self._widget -= self

	def move_up(self):
		"""
		Move the track up
		"""
		self._widget.move_track_up(self)

	def move_down(self):
		"""
		Move the track down
		"""
		self._widget.move_track_down(self)

	def draw_background(self, painter, start, end):
		painter.setOpacity(0.1)
		painter.fillRect(start, self.top_coordinate, end-start, self._widget.TRACK_HEIGHT, QtCore.Qt.black)
		painter.setOpacity(1.0)

	def draw(self, painter, start, end):
		"""
		
		:param painter: 
		:param start: 
		:param end: 
		:param index: 
		:return: 
		"""
		painter.drawLine(start, self.top_coordinate, end, self.top_coordinate)

	def draw_events(self, painter, start, end):
		"""
		
		:param painter: 
		:param start: 
		:param end: 
		:return: 
		"""
		for time in self._events:
			painter.setBrush(time.color)
			time.draw(painter)

	def draw_title(self, painter, start, end):
		"""
		
		:param painter: 
		:param index: 
		:return: 
		"""
		painter.setPen(QtCore.Qt.black)
		painter.setOpacity(0.5)
		painter.drawText(start+10, self.top_coordinate + painter.fontMetrics().height(), self.title)
		painter.setOpacity(1.0)

	def select_event(self, x, y):
		"""
		
		:param x: 
		:param y: 
		:return: 
		"""
		for delta in self._events:
			if delta.collide(x, y): return delta
		return None


	def clear(self):
		"""
		Clean all the events from the tracks.
		"""
		del self._events[:]
		self._events = []

	@property
	def properties(self):
		return ['T', self.title, self.color.name()]

	@property
	def top_coordinate(self):
		"""
		:return: Top pixel coordinate
		"""
		return self._index*self._widget.TRACK_HEIGHT + self._widget.TOPTRACK_HEIGHT

	@property
	def bottom_coordinate(self):
		"""
		:return: Bottom pixel coordinate
		"""
		return self.top_coordinate + self._widget.TRACK_HEIGHT



#!/usr/bin/python
# -*- coding: utf-8 -*-

import os
from pyforms_gui.controls.control_base import ControlBase
from AnyQt           import uic
from AnyQt.QtWidgets import QApplication

class ControlProgress(ControlBase):

    def __init__(self, *args, **kwargs):
        self._updateSlider = True
        self._min = kwargs.get('minimum', 0)
        self._max = kwargs.get('maximum', 100)
        if 'default' not in kwargs: kwargs['default'] = 0
        ControlBase.__init__(self, *args, **kwargs)

        

    def init_form(self):
        module_path = os.path.abspath(os.path.dirname(__file__))
        control_path = os.path.join(module_path, "progressInput.ui")
        self._form = uic.loadUi(control_path)
        self._form.horizontalSlider.setMinimum(self._min)
        self._form.horizontalSlider.setMaximum(self._max)
        self._form.horizontalSlider.setValue(self._value)
        self._form.horizontalSlider.setFormat(self._label)
        super(ControlProgress, self).init_form()

    @property
    def label(self): return self._label

    @label.setter
    def label(self, value):
        self._label = value
        self._form.horizontalSlider.setFormat(self._label)

    @property
    def value(self): return self._form.horizontalSlider.value()

    @value.setter
    def value(self, value):
        self._form.horizontalSlider.setValue(value)
        QApplication.processEvents()

    def __add__(self, other):
        self.value = self.value + other
        return self

    def __sub__(self, other):
        self.value = self.value - other
        return self

    @property
    def min(self): return self._form.horizontalSlider.minimum()

    @min.setter
    def min(self, value): self._form.horizontalSlider.setMinimum(value)

    @property
    def max(self): return self._form.horizontalSlider.maximum()

    @max.setter
    def max(self, value): self._form.horizontalSlider.setMaximum(value)



#!/usr/bin/python
# -*- coding: utf-8 -*-

__author__ = "Ricardo Ribeiro"
__credits__ = ["Ricardo Ribeiro"]
__license__ = "MIT"
__version__ = "0.0"
__maintainer__ = "Ricardo Ribeiro"
__email__ = "ricardojvr@gmail.com"
__status__ = "Development"

import pyforms_gui.utils.tools as tools
from AnyQt 		import uic, QtGui
from pyforms_gui.controls.control_base import ControlBase

class ControlTextArea(ControlBase):

	def __init__(self, *args, **kwargs):
		ControlBase.__init__(self, *args, **kwargs)
		self.autoscroll = kwargs.get('autoscroll', False)
		
		

	def init_form(self):
		control_path = tools.getFileInSameDirectory(__file__, "textArea.ui")
		self._form = uic.loadUi(control_path)
		self._form.label.setText(self._label)
		if self._value:
			self._form.plainTextEdit.setPlainText(str(self._value))

		if not self._label or len(self._label)==0: 
			self.form.label.hide()

		super(ControlTextArea, self).init_form()
		self.form.plainTextEdit.textChanged.connect(self.finishEditing)

	def __add__(self, other):
		self._form.plainTextEdit.appendPlainText(str(other))

		# if activated the text field is autoscrolled to the bottom
		if self.autoscroll:
			self._form.plainTextEdit.moveCursor(QtGui.QTextCursor.End)

		return self

	def finishEditing(self):
		"""Function called when the lineEdit widget is edited"""
		self.changed_event()
		

	@property
	def value(self):
		return self._form.plainTextEdit.toPlainText()

	@value.setter
	def value(self, value):
		self._form.plainTextEdit.setPlainText(str(value))

	@property
	def readonly(self):
		return self._form.plainTextEdit.isReadOnly()

	@readonly.setter
	def readonly(self, value):
		self._form.plainTextEdit.setReadOnly(value)


	@property
	def autoscroll(self): return self._autoscroll
	@autoscroll.setter
	def autoscroll(self, value): self._autoscroll = value



#!/usr/bin/python
# -*- coding: utf-8 -*-

""" pyforms_gui.controls.ControlList

"""
import logging
import os

from confapp import conf

from AnyQt           import QtCore, uic
from AnyQt.QtWidgets import QTableView, QAbstractItemView
from AnyQt.QtGui     import QIcon

from pyforms_gui.basewidget import BaseWidget
from pyforms_gui.controls.control_base import ControlBase

logger = logging.getLogger(__name__)


class ControlTableView(ControlBase, QTableView):
    """ This class represents a wrapper to the table widget
        It allows to implement a list view
    """

    CELL_VALUE_BEFORE_CHANGE = None  # store value when cell is double clicked

    def __init__(self, *args, **kwargs):
        QTableView.__init__(self)
        ControlBase.__init__(self, *args, **kwargs)

        if kwargs.get('select_entire_row', False):
            self.setSelectionBehavior(QAbstractItemView.SelectRows)

    
    ##########################################################################
    ############ EVENTS ######################################################
    ##########################################################################

    
    ##########################################################################
    ############ PROPERTIES ##################################################
    ##########################################################################



    @property
    def value(self):
        return self.model()
        
    @value.setter
    def value(self, value):
        self.setModel(value)
  
    

   
    @property
    def form(self):
        return self

    
    ##########################################################################
    ############ PRIVATE FUNCTIONS ###########################################
    ##########################################################################



# !/usr/bin/python
# -*- coding: utf-8 -*-

from pyforms_gui.controls.control_base import ControlBase
from AnyQt.QtWidgets import QTreeWidget, QTreeWidgetItem, QTreeView, QAbstractItemView, QAction
from AnyQt.QtGui import QIcon, QKeySequence
from AnyQt import QtCore

class ControlTree(ControlBase, QTreeWidget):
    """This class represents a wrapper to the QTreeWidget"""

    def __init__(self, *args, **kwargs):
        QTreeWidget.__init__(self)
        ControlBase.__init__(self, *args, **kwargs)

    def init_form(self):
        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.setUniformRowHeights(True)
        self.setDragDropMode(QAbstractItemView.NoDragDrop)
        self.setDragEnabled(False)
        self.setAcceptDrops(False)

        self.model().dataChanged.connect(self.__itemChangedEvent)
        self.itemDoubleClicked.connect(self.__itemDoubleClicked)

        self.selectionChanged = self.selectionChanged
        self._items = {}
        super(ControlTree, self).init_form()

    def __repr__(self):
        return QTreeWidget.__repr__(self)

    ##########################################################################
    ############ FUNCTIONS ###################################################
    ##########################################################################

    def __add__(self, other):
        if isinstance(other, QTreeWidgetItem):
            self.invisibleRootItem().addChild(other)
        elif isinstance(other, str):
            item = QTreeWidgetItem(other)
            self.invisibleRootItem().addChild(item)
        elif isinstance(other, list):
            for x in other:
                if isinstance(x, str):
                    item = QTreeWidgetItem(x)
                    self.invisibleRootItem().addChild(item)
                else:
                    self.invisibleRootItem().addChild(x)
        else:
            item = QTreeWidgetItem(other)
            self.invisibleRootItem().addChild(item)

        # self.setFirstColumnSpanned( self.model().rowCount() - 1, self.rootIndex(), True)
        return self

    def __remove_recursively(self, parent, item_2_remove):
        if parent is None: return

        for i in range(parent.childCount()):
            child = parent.child(i)
            if child == item_2_remove:
                parent.removeChild(child)
            else:
                self.__remove_recursively(child, item_2_remove)

    def __sub__(self, other):
        if isinstance(other, int):
            if other < 0:
                indexToRemove = self.selected_row_index
            else:
                indexToRemove = other
            self.model().removeRow(indexToRemove)
        else:
            self.__remove_recursively(self.invisibleRootItem(), other)
        return self

    def save_form(self, data, path=None):
        pass

    def load_form(self, data, path=None):
        pass

    def add_popup_menu_option(self, label='', function_action=None, key=None, item=None, icon=None, submenu=None):
        """
        Add an option to the Control popup menu
        @param label:           label of the option.
        @param function_action:  function called when the option is selected.
        @param key:             shortcut key
        @param key:             shortcut key
        """
        action = super(ControlTree, self).add_popup_menu_option(label, function_action, key, submenu)

        if item is not None:

            if label == "-":
                self._items[id(item)].append(label)
            else:
                action = QAction(label, self.form)
                if icon is not None:
                    action.setIconVisibleInMenu(True)
                    action.setIcon(QIcon(icon))
                if key is not None:
                    action.setShortcut(QKeySequence(key))
                if function_action:
                    action.triggered.connect(function_action)
                    # Associate action to the item.
                    if id(item) not in self._items.keys():
                        self._items.update({id(item): []})
                    self._items[id(item)].append(action)
                ##########################
                return action
        return action

    def clear(self):
        super(ControlTree, self).clear()
        if self._popup_menu:
            self._popup_menu.clear()
        self._items = {}

    def expand_item(self, item, expand=True, parents=True):
        item.setExpanded(expand)
        if parents:
            parent = item.parent()
            while (True):
                try:
                    parent.setExpanded(expand)
                    parent = parent.parent()
                except AttributeError:
                    break

    def create_child(self, name, parent=None, icon=None):
        """
        Create a new child for to the parent item.
        If the parent is None it add to the root.
        """
        item = QTreeWidgetItem(self, [name]) if (
            parent is None) else QTreeWidgetItem(parent, [name])
        if icon is not None:
            if isinstance(icon, str):
                item.setIcon(0, QIcon(icon))
            elif isinstance(icon, QIcon):
                item.setIcon(0, icon)
        return item

    ##########################################################################
    ############ EVENTS ######################################################
    ##########################################################################

    def item_changed_event(self, item):
        pass

    def item_selection_changed_event(self):
        pass

    def item_double_clicked_event(self, item):
        pass

    def key_press_event(self, event):
        pass

    def rows_inserted_event(self, parent, start, end):
        """ This event is called every time a new row is added to the tree"""
        pass

    ##########################################################################
    ############ PROPERTIES ##################################################
    ##########################################################################

    @property
    def show_header(self):
        return self.header().isVisible()

    @show_header.setter
    def show_header(self, value):
        self.header().show() if value else self.header().hide()

    @property
    def selected_rows_indexes(self):
        result = []
        for index in self.selectedIndexes():
            result.append(index.row())
        return list(set(result))

    @property
    def selected_row_index(self):
        indexes = self.selected_rows_indexes
        if len(indexes) > 0:
            return indexes[0]
        else:
            return None

    @selected_row_index.setter
    def selected_row_index(self, value):
        self.setCurrentCell(value)

    @property
    def selected_item(self):
        return self.selectedItems()[0] if len(self.selectedItems()) > 0 else None

    @selected_item.setter
    def selected_item(self, value):
        self.setCurrentItem(value)

    @property
    def form(self):
        return self

    @property
    def value(self):
        root = self.invisibleRootItem()
        return [root.child(i) for i in range(root.childCount())]

    @value.setter
    def value(self, value):
        if isinstance(value, list):
            for x in value:
                self += x
        else:
            self += value

    @property
    def icon_size(self):
        size = self.iconSize()
        return size.width(), size.height()

    @icon_size.setter
    def icon_size(self, value):
        self.setIconSize(QtCore.QSize(*value))

    ##########################################################################
    ############ PRIVATE FUNCTIONS ###########################################
    ##########################################################################

    def __itemChangedEvent(self, item):
        self.item_changed_event(item)

    def rowsInserted(self, parent, start, end):
        super(ControlTree, self).rowsInserted(parent, start, end)
        self.rows_inserted_event(parent, start, end)

    def selectionChanged(self, selected, deselected):
        super(QTreeView, self).selectionChanged(selected, deselected)
        self.item_selection_changed_event()

    def __itemDoubleClicked(self, item, column):
        if hasattr(item, 'double_clicked_event'): item.double_clicked_event()
        self.item_double_clicked_event(item)

    def keyPressEvent(self, event):
        QTreeView.keyPressEvent(self, event)
        item = self.selected_item
        if hasattr(item, 'key_pressed_event'): item.key_pressed_event(event)
        self.key_press_event(event)

    def about_to_show_contextmenu_event(self):
        """
        Function called before open the Control popup menu
        """
        if len(self._items) > 0:  # Reset the menu and construct a new one only if there are actions for the items.
            self._popup_menu.clear()
            itemSelected = self.selectedItems()[0]

            if id(itemSelected) in self._items:
                for action in self._items[id(itemSelected)]:
                    if action == '-':
                        self._popup_menu.addSeparator()
                    else:
                        self._popup_menu.addAction(action)
                    # print("Adding action {action} to {item}".format(
                    #    action=action.text(), item=itemSelected))

    def clone_item(self, parent, item, copy_function=None):
        new_item = QTreeWidgetItem()
        for col_index in range(item.columnCount()):
            new_item.setText(col_index, item.text(col_index))
            new_item.setIcon(col_index, item.icon(col_index))
        if copy_function is not None: copy_function(item, new_item)
        parent.addChild(new_item)
        for child_index in range(item.childCount()):
            child_item = item.child(child_index)
            self.clone_item(new_item, child_item, copy_function)

    def clone_tree(self, tree, copy_function=None):

        for item in tree.value:
            self.clone_item(self.invisibleRootItem(), item, copy_function)



#!/usr/bin/python
# -*- coding: utf-8 -*-


import pyforms_gui.utils.tools as tools

from confapp import conf

from AnyQt.QtWidgets import QLabel, QWidget, QComboBox, QHBoxLayout, QSizePolicy

from pyforms_gui.controls.control_base import ControlBase

class ValueNotSet: pass

class ControlCombo(ControlBase, QWidget):
    """This class represents a wrapper to the combo box"""

    def __init__(self, *args, **kwargs):        
        QWidget.__init__(self)
        ControlBase.__init__(self, *args, **kwargs)

        for item in kwargs.get('items', []):
            self.add_item(*item);
        
    ##########################################################################
    ############ Functions ###################################################
    ##########################################################################

    def init_form(self):

        self._layout = QHBoxLayout()
        self._combo  = QComboBox(self.form)
        
        if self._label is not None:
            self._combolabel = QLabel(self.form)
            self._layout.addWidget( self._combolabel )
            self._combolabel.setAccessibleName('ControlCombo-label')
            self.label = self._label
        else:
            self._combolabel = None

        self._layout.addWidget( self._combo )
        self.form.setLayout( self._layout )


        self._combo.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        self._layout.setContentsMargins(0,0,0,0)
        self.form.setContentsMargins(0,0,0,0)
        self.form.setMinimumHeight(38)
        self.form.setMaximumHeight(38)
        self.form.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        
        self._combo.currentIndexChanged.connect(self._currentIndexChanged)
        self._combo.activated.connect(self._activated)
        self._combo.highlighted.connect(self._highlighted)
        self._combo.editTextChanged.connect(self._editTextChanged)
        
        self._items = {}
        self._addingItem = False
        super(ControlCombo, self).init_form()

        

    def clear(self):
        self._items = {}
        self._value = None
        self._combo.clear()

    def add_item(self, label, value=ValueNotSet):
        self._addingItem = True
        
        if value is not ValueNotSet:
            if not (value in self._items.values()):
                self._combo.addItem(label)
        else:
            if not (label in self._items.keys()):
                self._combo.addItem(label)

        firstValue = False
        if self._items == {}:
            firstValue = True

        if value is ValueNotSet:
            self._items[str(label)] = label
        else:
            self._items[str(label)] = value
        self._addingItem = False

        if firstValue:
            self.value = self._items[label]

    def __add__(self, val):
        if isinstance(val, tuple):
            self.add_item(val[0], val[1])
        else:
            self.add_item(val)

        return self

    def get_item_index_by_name(self, item_name):
        """
        Returns the index of the item containing the given name
        :param item_name: item name in combo box
        :type item_name: string
        """
        return self._combo.findText(item_name)

    def count(self):
        return self._combo.count()

    def show(self):
        """
        Show the control
        """
        QWidget.show(self)

    def hide(self):
        """
        Hide the control
        """
        QWidget.hide(self)


    ##########################################################################
    ############ Events ######################################################
    ##########################################################################

    def current_index_changed_event(self, index):
        """Called when the user chooses an item in the combobox and
        the selected choice is different from the last one selected.
        @index: item's index
        """
        pass

    def activated_event(self, index):
        """Called when the user chooses an item in the combobox.
        Note that this signal happens even when the choice is not changed
        @index: item's index
        """
        pass

    def highlighted_event(self, index):
        pass

    def edittext_changed_event(self, text):
        pass

    ##########################################################################
    ############ PROPERTIES ##################################################
    ##########################################################################

    @property
    def form(self):
        return self

    @property
    def current_index(self):
        return self._combo.currentIndex()

    @current_index.setter
    def current_index(self, value):
        self._combo.setCurrentIndex(value)

    @property
    def values(self):
        return self._items.values()

    @property
    def items(self):
        return self._items.items()

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        for key, val in self.items:
            if value == val:
                index = self._combo.findText(key)
                self._combo.setCurrentIndex(index)
                if self._value != value:
                    self.changed_event()
                self._value = val

    @property
    def text(self):
        return str(self._combo.currentText())

    @text.setter
    def text(self, value):
        for key, val in self.items:
            if value == key:
                self.value = val
                break

    @property
    def label(self):
        if self._combolabel:
            return self._combolabel.text()
        else:
            return None

    @label.setter
    def label(self, value):
        """
        Label of the control, if applies
        @type  value: string
        """
        if self._combolabel:
            self._combolabel.setText(value)
        

    ##########################################################################
    ############ Private functions ###########################################
    ##########################################################################

    def _activated(self, index):
        if not self._addingItem:
            item = self._combo.currentText()
            if len(item) >= 1:
                ControlBase.value.fset(self, self._items[str(item)])
                self.activated_event(index)

    def _highlighted(self, index):
        """Called when an item in the combobox popup
         list is highlighted by the user.
         @index: item's index
        """
        self.highlighted_event(index)

    def _editTextChanged(self, text):
        self.edittext_changed_event(text)

    def _currentIndexChanged(self, index):
        if not self._addingItem:
            item = self._combo.currentText()
            item = str(item)

            if item in self._items:
                ControlBase.value.fset(self, self._items[ str(item) ] )
                self.current_index_changed_event(index)



#!/usr/bin/python
# -*- coding: utf-8 -*-

import csv, os
from confapp import conf
from AnyQt import QtCore, _api
from AnyQt.QtWidgets import QWidget, QScrollArea, QFileDialog, QMessageBox, QPushButton, QLabel, QSlider, QHBoxLayout, QVBoxLayout
from pyforms_gui.controls.control_base import ControlBase
from pyforms_gui.controls.control_event_timeline.utils.import_window import ImportWindow
from pyforms_gui.controls.control_event_timeline.timeline_widget import TimelineWidget
from pyforms_gui.controls.control_event_timeline.events.win_track import TimelinePopupWindow
from pyforms_gui.controls.control_event_timeline.graphs.graph import Graph
from pyforms_gui.controls.control_event_timeline.graphs.win_graph_to_event import Graph2Event
from pyforms_gui.controls.control_event_timeline.graphs.win_graph_properties  import GraphsProperties
from pyforms_gui.controls.control_event_timeline.graphs.win_events_generator import GraphsEventsGenerator
import traceback


class ControlEventTimeline(ControlBase, QWidget):
    """
        Timeline events editor

        **Short keys:**

            - **Control + Left**: Move event to the left.
            - **Control + Right**: Move event to the right.
            - **Delete**: Delete an event.
            - **L**: Lock an event.
            - **Control + Up**: Move an event up.
            - **Control + Down**: Move an event down.
            - **Shift + Control + Left**: Move an event end time to the left.
            - **Shift + Control + Right**: Move an event end to the right.
            - **Shift + Left**: Move an event beginning to the left.
            - **Shift + Right**: Move an event beginning to the right.
            - **S**: First press, mark the beginning of an event, Second press, create an event ending in the current cursor time.
            - **A**: Move the cursor to the left.
            - **D**: Move the cursor to the right.
            - **Q**: Select the previous event in the selected row.
            - **E**: Select the next event in the selected row.
    """

    def __init__(self, label="", default=0, max=100):
        QWidget.__init__(self)
        ControlBase.__init__(self, label, default)
        self._max = 100
        self._graphs_prop_win     = GraphsProperties(self._time, self)
        self._graphsgenerator_win = GraphsEventsGenerator(self._time)
        self._graph2event_win     = Graph2Event(self._time)

        ###############################################################################################
        ######## EVENTS ACTIONS #######################################################################
        ###############################################################################################

        # Popup menus that only show when clicking on a TIMELINEDELTA object
        event_remove_action = self.add_popup_menu_option(
            "Remove event", self.__removeSelected, key='Delete',
            icon=conf.ANNOTATOR_ICON_DELETE
        )
        separator_action = self.add_popup_menu_option("-")
        self._events_actions = [event_remove_action, separator_action]
        for action in self._events_actions: action.setVisible(False)

        ###############################################################################################
        ######## TRACKS ACTIONS #######################################################################
        ###############################################################################################

        # General right click popup menus
        track_properties_action = self.add_popup_menu_option(
            "Row properties",
            self.__open_track_properties_evt,
            icon=conf.ANNOTATOR_ICON_INFO
        )

        track_insert_action = self.add_popup_menu_option(
            "Insert row",
            self.__add_track_evt,
            icon=conf.ANNOTATOR_ICON_ADD
        )

        track_remove_action = self.add_popup_menu_option(
            "Remove row",
            self.__remove_current_track_evt,
            icon=conf.ANNOTATOR_ICON_DELETE
        )

        track_moveup_action = self.add_popup_menu_option(
            "Move up",
            self.__move_track_up_evt,
            icon=conf.PYFORMS_ICON_EVENTTIMELINE_IMPORT
        )

        track_movedown_action = self.add_popup_menu_option(
            "Move down",
            self.__move_track_down_evt,
            icon=conf.PYFORMS_ICON_EVENTTIMELINE_EXPORT
        )

        separator_action = self.add_popup_menu_option("-")

        self._tracks_actions = [
            track_properties_action, track_insert_action,
            track_remove_action, track_moveup_action,
            track_movedown_action, separator_action
        ]
        for action in self._tracks_actions: action.setVisible(False)



        ###############################################################################################
        ######## GRAPHS ACTIONS #######################################################################
        ###############################################################################################

        self.add_popup_menu_option("Graphs", self.open_graphs_properties, icon=conf.PYFORMS_ICON_EVENTTIMELINE_GRAPH)
        self.add_popup_menu_option("Apply a function to the graphs", self.__generate_graphs_events,
                                   icon=conf.PYFORMS_ICON_EVENTTIMELINE_GRAPH)
        self.add_popup_menu_option("Convert graph to events", self.__graph2event_event,
                                   icon=conf.PYFORMS_ICON_EVENTTIMELINE_GRAPH)
        
        self.add_popup_menu_option("-")
        
        self.add_popup_menu_option("Auto adjust rows", self.__auto_adjust_tracks_evt,
                                   icon=conf.PYFORMS_ICON_EVENTTIMELINE_REFRESH)

        self.add_popup_menu_option(
            "Remove everything",
            self.clean,
            icon=conf.ANNOTATOR_ICON_DELETE
        )

        self._importwin = None # import window.

    def __repr__(self):
        return "Timeline "+str(self.name)

    def init_form(self):
        # Get the current path of the file
        rootPath = os.path.dirname(__file__)

        vlayout = QVBoxLayout()
        hlayout = QHBoxLayout()

        if _api.USED_API == _api.QT_API_PYQT5:
            hlayout.setContentsMargins(0,0,0,0)
            vlayout.setContentsMargins(0,0,0,0)
        elif _api.USED_API == _api.QT_API_PYQT4:
            hlayout.setMargin(0)
            vlayout.setMargin(0)

        self.setLayout(vlayout)

        # Add scroll area
        scrollarea = QScrollArea()
        self._scrollArea = scrollarea
        scrollarea.setMinimumHeight(140)
        scrollarea.setWidgetResizable(True)
        scrollarea.keyPressEvent = self.__scrollAreaKeyPressEvent
        scrollarea.keyReleaseEvent = self.__scrollAreaKeyReleaseEvent

        vlayout.addWidget(scrollarea)

        # The timeline widget
        self._time = widget = TimelineWidget(self)
        widget._scroll = scrollarea
        scrollarea.setWidget(widget)

        # Timeline zoom slider
        slider = QSlider(QtCore.Qt.Horizontal)
        slider.setFocusPolicy(QtCore.Qt.NoFocus)
        slider.setMinimum(1)
        slider.setMaximum(100)
        slider.setValue(10)
        slider.setPageStep(1)
        slider.setTickPosition(QSlider.NoTicks)  # TicksBothSides
        slider.valueChanged.connect(self.__scaleSliderChange)

        slider_label_zoom_in = QLabel()
        slider_label_zoom_out = QLabel()
        slider_label_zoom_in.setPixmap(conf.PYFORMS_PIXMAP_EVENTTIMELINE_ZOOM_IN)
        slider_label_zoom_out.setPixmap(conf.PYFORMS_PIXMAP_EVENTTIMELINE_ZOOM_OUT)

        self._zoomLabel = QLabel("100%")
        hlayout.addWidget(self._zoomLabel)
        hlayout.addWidget(slider_label_zoom_out)
        hlayout.addWidget(slider)
        hlayout.addWidget(slider_label_zoom_in)

        # Import/Export Buttons
        btn_import = QPushButton("Import")

        btn_import.setIcon(conf.PYFORMS_ICON_EVENTTIMELINE_IMPORT)
        btn_import.clicked.connect(self.__open_import_win_evt)
        btn_export = QPushButton("Export")

        btn_export.setIcon(conf.PYFORMS_ICON_EVENTTIMELINE_EXPORT)
        btn_export.clicked.connect(self.__export)
        hlayout.addWidget(btn_import)
        hlayout.addWidget(btn_export)

        vlayout.addLayout(hlayout)


    ##########################################################################
    #### HELPERS/PUBLIC FUNCTIONS ############################################
    ##########################################################################

    def __add__(self, other):
        if isinstance(other, Graph):
            self._graphs_prop_win     += other
            self._graphsgenerator_win += other
            self._graph2event_win     += other
        return self

    def __sub__(self, other):
        if isinstance(other, int): 
            self._graphs_prop_win     -= other
            self._graphsgenerator_win -= other
            self._graph2event_win     -= other
        return self

    def move_selected_event_or_pointer_left(self):
        """
        Move the selected event or pointer to the left.
        :return:
        """
        if self.timeline_widget.selected:
            self.timeline_widget.selected.move(-1, 0)
            self.timeline_widget.repaint()

    def move_selected_event_or_pointer_right(self):
        """
        Move the selected event or pointer to the right.
        :return:
        """
        if self.timeline_widget.selected:
            self.timeline_widget.selected.move(1, 0)
            self.timeline_widget.repaint()

    def remove_selected_event(self):
        """
        Remove the selected event.
        :return:
        """
        if self.timeline_widget.selected == self.timeline_widget.pointer or self.timeline_widget.selected is None:
            return

        self.timeline_widget.remove_selected_event()

    def toggle_selected_event_lock(self):
        """
        Toggle the lock of the selected event.
        :return:
        """
        if self.timeline_widget.selected == self.timeline_widget.pointer or self.timeline_widget.selected is None:
            return

        self.timeline_widget.toggle_selected_event_lock()


    def select_next_event(self):
        """
        Select the next event.
        :return:
        """
        if self.timeline_widget.selected == self.timeline_widget.pointer or self.timeline_widget.selected is None:
            return

        index = self.timeline_widget.selected_row.events.index(self.timeline_widget.selected)
        if index < len(self.timeline_widget.selected_row.events) - 1:
            self.timeline_widget.selected = self.timeline_widget.selected_row.events[index + 1]
            self.timeline_widget.position = self.timeline_widget.selected.begin

    def select_previous_event(self):
        """
        Select the previous event.
        :return:
        """
        if self.timeline_widget.selected == self.timeline_widget.pointer or self.timeline_widget.selected is None:
            return

        index = self.timeline_widget.selected_row.events.index(self.timeline_widget.selected)
        if index > 0:
            self.timeline_widget.selected = self.timeline_widget.selected_row.events[index - 1]
            self.timeline_widget.position = self.timeline_widget.selected.begin

    def select_first_event(self):
        """
        Select the first event.
        :return:
        """
        if self.timeline_widget.selected == self.timeline_widget.pointer or self.timeline_widget.selected is None:
            return

        if self.timeline_widget.selected_row is not None and len(
                self.timeline_widget.selected_row) > 0:
            self.timeline_widget.selected = self.timeline_widget.selected_row.events[0]
            self.timeline_widget.position = self.timeline_widget.selected.begin

    def select_last_event(self):
        """
        Select the last event.
        :return:
        """
        if self.timeline_widget.selected == self.timeline_widget.pointer or self.timeline_widget.selected is None:
            return

        if self.timeline_widget.selected_row is not None and len(
                self.timeline_widget.selected_row) > 0:
            self.timeline_widget.selected = self.timeline_widget.selected_row.events[
                len(self.timeline_widget.selected_row) - 1]
            self.timeline_widget.position = self.timeline_widget.selected.begin


    def move_selected_event_up(self):
        """
        Move the selected event to the track above.
        :return:
        """
        if self.timeline_widget.selected == self.timeline_widget.pointer or self.timeline_widget.selected is None:
            return

        self.timeline_widget.selected.move(
            0,
            self.timeline_widget.selected.top_coordinate - self.timeline_widget.TRACK_HEIGHT
        )
        self.timeline_widget.repaint()

    def move_selected_event_down(self):
        """
        Move the selected event to the track bellow.
        :return:
        """
        if self.timeline_widget.selected == self.timeline_widget.pointer or self.timeline_widget.selected is None:
            return

        self.timeline_widget.selected.move(
            0,
            self.timeline_widget.selected.top_coordinate + self.timeline_widget.TRACK_HEIGHT
        )
        self.timeline_widget.repaint()

    def move_selected_event_end_left(self):
        """
        Move the selected event end to the left.
        :return:
        """
        if self.timeline_widget.selected == self.timeline_widget.pointer or self.timeline_widget.selected is None:
            return

        self.timeline_widget.selected.move_end(-1)
        self.timeline_widget.repaint()

    def move_selected_event_end_right(self):
        """
        Move the selected event end to the right.
        :return:
        """
        if self.timeline_widget.selected == self.timeline_widget.pointer or self.timeline_widget.selected is None:
            return

        self.timeline_widget.selected.move_end(1)
        self.timeline_widget.repaint()

    def move_selected_event_begin_left(self):
        """
        Move the selected event begin to the left.
        :return:
        """
        if self.timeline_widget.selected == self.timeline_widget.pointer or self.timeline_widget.selected is None:
            return

        self.timeline_widget.selected.move_begin(-1)
        self.timeline_widget.repaint()

    def move_selected_event_begin_right(self):
        """
        Move the selected event begin to the right.
        :return:
        """
        if self.timeline_widget.selected == self.timeline_widget.pointer or self.timeline_widget.selected is None:
            return

        self.timeline_widget.selected.move_begin(1)
        self.timeline_widget.repaint()

    def open_and_close_event(self):
        """
        Open and close and event. The first time the function is called an event is opened.
        On the second call the event is closed and added to the timeline.
        :return:
        """
        self.timeline_widget.open_and_close_event()

    def add_event(self, begin, end, title='', row=0, track=None):
        """
        :param begin: Initial frame
        :param end: Last frame
        :param title: Event title
        :param row: Row to which the event should be added.
        """

        self._time.add_event(begin, end, title=title, track=track, row=row)
        self._time.repaint()

    def add_graph(self, name, data):
        """
        
        :param name: 
        :param data: 
        :return: 
        """
        self._time.add_graph(name, data)

    def add_track(self, title='', color=None):
        """
        Add a new track.
        :param str title: Title of the track.
        :param QColor color: Default color of the events in the track.
        :return: Return the added track.
        """
        return self._time.add_track(title=title, color=color)

    def get_track(self, title):
        """
        Get a track by its title
        :param str title: Title of the track.
        :return: Return the track with the matching title.
        """
        return self._time.get_track(title)


    def rename_graph(self, graph_index, newname):
        """
        Rename a graph by index.
        :param int graph_index: Index of the graph to rename.
        :param str newname: New name
        """
        self._time.graphs[graph_index].name = newname
        self._graphs_prop_win.rename_graph(graph_index, newname)
        self._graphsgenerator_win.rename_graph(graph_index, newname)
        self._graph2event_win.rename_graph(graph_index, newname)


    def import_graph_csv(self, filepath, separator=';', ignore_rows=0):
        """
        Import a new graph from a csv file.
        :param filename: 
        :param separator: 
        :param ignore_rows: 
        :return: 
        """
        with open(filepath, 'U') as csvfile:
            spamreader = csv.reader(csvfile, delimiter=separator)
            for i in range(ignore_rows): next(spamreader, None)
            name = os.path.basename(filepath).replace('.csv', '').replace('.CSV', '')
            chart = self._time.add_graph(name, spamreader)
            return chart

    def export_csv_file(self, filename):
        with open(filename, 'w') as csvfile:
            spamwriter = csv.writer(csvfile, dialect='excel')
            self._time.export_events_to_csvwriter(spamwriter)

    def import_csv_file(self, filename):
        with open(filename, 'r') as csvfile:
            spamreader = csv.reader(csvfile, dialect='excel')
            self._time.import_events_from_csvreader(spamreader)

    def import_csv(self, csvfile):
        """

        :param csvfile:
        """
        # If there are annotation in the timeline, show a warning
        if len(self._time._tracks) > 0:  # dict returns True if not empty
            message = ["You are about to import new data. ",
                       "If you proceed, current annotations will be erased. ",
                       "Make sure to export current annotations first to save.",
                       "\n",
                       "Are you sure you want to proceed?"]
            reply = QMessageBox.question(self,
                                         "Warning!",
                                         "".join(message),
                                         QMessageBox.Yes | QMessageBox.No,
                                         QMessageBox.No)

            if reply != QMessageBox.Yes:
                return

        self._time.import_events_from_csvreader(csvfile)

    def open_graphs_properties(self):
        """
        Opens the graphs properties.
        """
        self._graphs_prop_win.show()
        self._time.repaint()


    def open_import_graph_win(self, filepath, frame_col=0, val_col=1):
        """
        Open a window to import a graph from a csv file.
        :param str filepath: Path of the file to import.
        :param int frame_col: Column corresponding to the frames number in the csv file.
        :param int val_col: Column corresponding to the values in the csv file.
        """
        self.__open_import_win_evt()
        self._importwin.import_chart(filepath, frame_col, val_col)


    ##########################################################################
    #### EVENTS ##############################################################
    ##########################################################################

    def mouse_moveover_timeline_event(self, event):
        self._graphs_prop_win.mouse_moveover_timeline_event(event)



    @property
    def pointer_changed_event(self):
        return self._time._pointer.moveEvent

    @pointer_changed_event.setter
    def pointer_changed_event(self, value):
        self._time._pointer.moveEvent = value

    def __auto_adjust_tracks_evt(self):
        for i in range(len(self._time.tracks) - 1, -1, -1):
            track = self._time.tracks[i]
            if len(track) == 0:
                self._time -= track
            else:
                break

    def __add_track_evt(self):
        self._time.add_track()

    def __remove_track_from_bottom_evt(self):
        if len(self._time.tracks) > 0:
            track = self._time.tracks[-1]
            if len(track) == 0:
                self._time -= track

    def __move_track_up_evt(self):
        self._time.selected_row.move_up()

    def __move_track_down_evt(self):
        self._time.selected_row.move_down()

    def __generate_graphs_events(self):
        self._graphsgenerator_win.show()

    def __graph2event_event(self):
        self._graph2event_win.show()

    ##########################################################################
    #### PROPERTIES ##########################################################
    ##########################################################################

    @property
    def timeline_widget(self):
        return self._time

    @property
    def value(self):
        return self._time.position

    @value.setter
    def value(self, value):
        ControlBase.value.fset(self, value)
        self._time.position = value

    @property
    def max(self):
        return self._time.minimumWidth()

    @max.setter
    def max(self, value):
        self._max = value
        self._time.setMinimumWidth(value)
        self.repaint()

    @property
    def form(self):
        return self

    @property
    def rows(self):
        return self._time.tracks

    @property
    def graphs(self): return self._time.graphs

    @property
    def key_release_event(self):
        return self._time.key_release_event
    @key_release_event.setter
    def key_release_event(self, value):
        self._time.key_release_event = value

    ##########################################################################
    #### PRIVATE FUNCTIONS ###################################################
    ##########################################################################

    def about_to_show_contextmenu_event(self):
        """
        Hide and show context menu options.
        """
        # Hide and show events actions.
        for action in self._events_actions:
            action.setVisible(
                True if self._time._selected is not None else False
            )
        # Hide and show tracks actions.
        for action in self._tracks_actions:
            action.setVisible(
                True if self._time._selected_track is not None else False
            )

    

    def __open_track_properties_evt(self):
        """
        This controls makes possible the edition of a track in the
        timeline, based on the position of the mouse.

        Updates:
        - Track label
        - Track default color
        """
        current_track = self._time.current_mouseover_track
        parent = self._time

        # Tracks info dict and index
        i = current_track

        # Save current default color to override with selected track color
        timeline_default_color = parent.color
        try:
            parent.color = self._time._tracks[current_track].color
        except Exception as e:
            error_message = ("You tried to edit an empty track.",
                             "\n",
                             "Initialize it by creating an event first.")
            QMessageBox.warning(
                parent, "Attention!", "".join(error_message))
            return e

        # Create dialog
        dialog = TimelinePopupWindow(parent, i)
        dialog.setModal(True)  # to disable main application window

        # If dialog is accepted, update dict info
        if dialog._ui.exec_() == dialog.Accepted:
            # Update label
            if dialog.behavior is not None:
                self._time._tracks[i].title = dialog.behavior

            # Update color
            if self._time._tracks[i].color != dialog.color:
                for delta in self._time._tracks[i].events:
                    delta.color = dialog.color
                self._time._tracks[i].color = dialog.color
            self._time.repaint()
        else:
            pass

        # Restore timeline default color
        parent.color = timeline_default_color

    def __lockSelected(self):
        self._time.toggle_selected_event_lock()

    def __removeSelected(self):
        self._time.remove_selected_event()

    def __open_import_win_evt(self):
        """Import annotations from a file."""
        if self._importwin is None:
            self._importwin = ImportWindow(self)
        self._importwin.show()

    def __export(self):
        """Export annotations to a file."""

        try:

            if conf.PYFORMS_DIALOGS_OPTIONS:
                filename, ffilter = QFileDialog.getSaveFileName(parent=self,
                     caption="Export annotations file",
                     directory="untitled.csv",
                     filter="CSV Files (*.csv);;CSV Matrix Files (*.csv)",
                     options=conf.PYFORMS_DIALOGS_OPTIONS)
            else:
                filename, ffilter = QFileDialog.getSaveFileName(parent=self,
                                                                caption="Export annotations file",
                                                                directory="untitled.csv",
                                                                filter="CSV Files (*.csv);;CSV Matrix Files (*.csv)")

            filename = str(filename)
            ffilter = str(ffilter)
            if filename != "":
                with open(filename, 'w') as csvfile:
                    spamwriter = csv.writer(csvfile, dialect='excel')
                    if ffilter == 'CSV Files (*.csv)':
                        self._time.export_events_to_csvwriter(spamwriter)
                    elif ffilter == 'CSV Matrix Files (*.csv)':
                        self._time.exportmatrix_events_to_csvwriter(spamwriter)

        except Exception as e:
            traceback.print_exc()
            m = QMessageBox(QMessageBox.Critical, 'Error', str(e))
            m.exec_()

    def __export_2_csv_matrix(self):
        try:
            QMessageBox.warning(
                self, "Important!", 'Please note that this file cannot be imported after.')

            filename, _ = QFileDialog.getSaveFileName(parent=self,
                                                   caption="Export matrix file",
                                                   directory="untitled.csv",
                                                   filter="CSV Files (*.csv)",
                                                   options=QFileDialog.DontUseNativeDialog)
            if filename != "":
                with open(filename, 'w') as csvfile:
                    spamwriter = csv.writer(csvfile, dialect='excel')
                    self._time.exportmatrix_events_to_csvwriter(spamwriter)
        except Exception as e:
            traceback.print_exc()
            m = QMessageBox(QMessageBox.Critical, 'Error', str(e))
            m.exec_()

    def __remove_current_track_evt(self):
        reply = QMessageBox.question(self, 'Confirm',
                                     "Are you sure you want to remove the row and its events?",
                                     QMessageBox.Yes |
                                     QMessageBox.No, QMessageBox.No)
        if reply == QMessageBox.Yes:
            self._time.remove_selected_track()


    def clean(self):
        reply = QMessageBox.question(self, 'Confirm',
                                     "Are you sure you want to clean everything?", QMessageBox.Yes |
                                     QMessageBox.No, QMessageBox.No)
        if reply == QMessageBox.Yes:
            self._time.clean()


    def __scaleSliderChange(self, value):
        scale = 0.1 * value
        self._time.setMinimumWidth(scale * self._max)
        self._time.scale = scale
        self._zoomLabel.setText(str(value * 10).zfill(3) + "%")

    def __scrollAreaKeyReleaseEvent(self, event):
        modifiers = int(event.modifiers())
        self._time.keyReleaseEvent(event)

        QScrollArea.keyReleaseEvent(self._scrollArea, event)

    def __scrollAreaKeyPressEvent(self, event):
        modifiers = event.modifiers()

        if modifiers & QtCore.Qt.ControlModifier:
            event.ignore()

        if modifiers &  QtCore.Qt.ShiftModifier:
            event.ignore()

        if modifiers &  QtCore.Qt.AltModifier:
            event.ignore()

        if event.isAccepted():
            QScrollArea.keyPressEvent(self._scrollArea, event)



#!/usr/bin/python
# -*- coding: utf-8 -*-

__author__ = "Ricardo Ribeiro"
__credits__ = ["Ricardo Ribeiro"]
__license__ = "MIT"
__version__ = "0.0"
__maintainer__ = "Ricardo Ribeiro"
__email__ = "ricardojvr@gmail.com"
__status__ = "Development"

from AnyQt.QtWebEngineWidgets 			import QWebEngineView as QWebView
from AnyQt.QtCore 						import QUrl
from AnyQt.QtWidgets                    import QSizePolicy
from pyforms_gui.controls.control_base 	import ControlBase


class ControlWeb(ControlBase, QWebView):
    def __init__(self, *args, **kwargs):
        QWebView.__init__(self)
        if 'load_finished_event' in kwargs: 
            self.load_finished_event=kwargs['load_finished_event']
        self.loadFinished.connect(self.__load_finished_evt)
        ControlBase.__init__(self, *args, **kwargs)

        
    def init_form(self):
        """
        Load Control and initiate the events
        """
        ControlBase.init_form(self)
        if self._value: QWebView.load(self, QUrl(self._value))
        if self.help: self.form.setToolTip(self.help)

        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding);
    ############################################################################
    ############ Properties ####################################################
    ############################################################################

    def __load_finished_evt(self, ok):
        self.load_finished_event(ok)

    def load_finished_event(self,ok):
        pass

    @property
    def value(self): return ControlBase.value.fget(self)

    @value.setter
    def value(self, value):
        ControlBase.value.fset(self, value)
        QWebView.load(self, QUrl(value))

    @property
    def html(self): return None

    @html.setter
    def html(self, value):
        QWebView.setHtml(self, value)

    @property
    def form(self):
        return self



# !/usr/bin/python
# -*- coding: utf-8 -*-
from pyforms_gui.controls.control_base import ControlBase

from AnyQt.QtWidgets import QTreeView, QAbstractItemView
from AnyQt.QtGui     import QStandardItem, QStandardItemModel


class ControlTreeView(ControlBase, QTreeView):

    default_width = None
    
    def __init__(self, *args, **kwargs):
        QTreeView.__init__(self)
        ControlBase.__init__(self, *args, **kwargs)

        self.item_selection_changed_event = kwargs.get('item_selection_changed_event', self.item_selection_changed_event)
        self.item_double_clicked_event    = kwargs.get('item_double_clicked_event',    self.item_double_clicked_event)

    def init_form(self):
        #self.header().hide()
        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.setUniformRowHeights(True)
        #self.setDragDropMode(QAbstractItemView.InternalMove)
        #self.setDragEnabled(True)
        #self.setAcceptDrops(True)

        self.value = QStandardItemModel() if self._value is None else self._value
        
        #self.selectionChanged       = self.__selectionChanged
        self.mouseDoubleClickEvent  = self.__itemDoubleClicked

        self.selectionModel().selectionChanged.connect(self.__selectionChanged)
        super(ControlTreeView, self).init_form()

    def item_selection_changed_event(self, selected, deselected):
        pass

    def item_double_clicked_event(self, evt):
        pass

    ###################################################
    ## OTHER FUNCTIONS ################################
    ###################################################

    def __itemDoubleClicked(self, evt):
        self.item_double_clicked_event(evt)

    def __selectionChanged(self, selected, deselected):
        #def selectionChanged(self, selected, deselected):
        self.item_selection_changed_event(selected, deselected)
        #super(QTreeView, self.form).selectionChanged(selected, deselected)
        
    ###################################################
    ## PROPERTIES #####################################
    ###################################################

    @property
    def selected_row_index(self):
        for index in self.selectedIndexes():
            return index.row()
        return None

    @property
    def selected_item(self):
        for index in self.selectedIndexes():
            res = []
            while index.isValid():
                res.append( index.data() )
                index = index.parent()
            return list(reversed(res))
        return None

    @property
    def value(self):
        return self.model()
        
    @value.setter
    def value(self, value): self.setModel(value)
    
    
    @property
    def form(self): return self
